/*
 *   Copyright (c) 1999 David Parsons. All rights reserved.
 *   
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *  3. All advertising materials mentioning features or use of this
 *     software must display the following acknowledgement:
 *     
 *   This product includes software developed by David Parsons
 *   (orc@pell.chi.il.us)
 *
 *  4. My name may not be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *     
 *  THIS SOFTWARE IS PROVIDED BY DAVID PARSONS ``AS IS'' AND ANY
 *  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL DAVID
 *  PARSONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 *  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *  THE POSSIBILITY OF SUCH DAMAGE.
 */


/*
 * par: partition table handling.
 */


#include <stdio.h>
#include <malloc.h>
#include <basis/rpopen.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <ndialog.h>

#include "util.h"
#include "pathnames.h"
#include "part.h"
#include "media.h"


/* IS_OK() picks disks that exist and are not either CDROMs or removable
 */
#define IS_OK(x)	( (x & ID_EXIST) && !(x & (ID_CDROM|ID_REMOVABLE)) )


/*
 * get_disk_par() gets the partition table for a given disk (not needed
 *                on a system with a working devfs)
 */
void
get_disk_par(char *device, int *nr_arry, struct raw_par **arry)
{
    struct raw_par curpar;
    FILE *pipe;
    char bfr[200];
    char *p;
    static char *sfdiskargs[] = { "sfdisk", "-dd", 0/* device slot */, 0 };

    sfdiskargs[2] = device;

    if (*arry == 0)
	*arry = malloc(1);

    if ( (pipe = rpopen(PATH_SFDISK, sfdiskargs, "r")) != 0) {
	while (fgets(bfr, sizeof bfr, pipe)) {
	    strtok(bfr, "\n");
	    if (bfr[0] != '/' || bfr[0] == 0)
		continue;
	    memset(&curpar, 0, sizeof curpar);
	    curpar.par_dev = strtok(bfr, ":");	/* get the device */

	    /* after the device comes a list of comma-separated
	     * identifiers
	     */
	    while ((p = strtok(0, ",")) != 0)
		if (sscanf(p, " size= %ld", &curpar.par_blocks) == 1) {
		    /* loadandgo wants the blocksize to be in 1k blocks */
		    curpar.par_blocks /= 2;
		}
		else if (strstr(p, "bootable") != 0)
		    curpar.par_bootflag = 1;
		else {
		    sscanf(p, " Id= %x", &curpar.par_fstype);
		    sscanf(p, " lastcyl= %d", &curpar.par_lastcyl);
		}
	    if (curpar.par_blocks == 0 || curpar.par_fstype == 0)
		continue;

	    /* if everything else is okay, strip trailing spaces
	     * from the partition name
	     */
	    strtok(curpar.par_dev, " ");

	    fprintf(stderr, "%s: %ld blocks, lastcyl=%d, type=%x\n",
			curpar.par_dev, curpar.par_blocks,
			curpar.par_lastcyl, curpar.par_fstype);
	    (*arry) = realloc(*arry, (1+*nr_arry)*sizeof (*arry)[0]);
	    (*arry)[*nr_arry] = curpar;
	    (*arry)[*nr_arry].par_dev = strdup(curpar.par_dev);
	    (*nr_arry)++;
	}
	rpclose(pipe);
    }
} /* get_disk_par */


/*
 * get_par_list(): gets a list of all the partitions on disks in a disk list.
 */
void
get_par_list(LIA disk_list, int *nr_par, struct raw_par **par)
{
    int x;

    *par = 0;
    *nr_par = 0;
    for (x=0; x < LIAcount(disk_list); x++)
	get_disk_par(LIAlist(disk_list)[x].id, nr_par, par);
} /* get_par_list */


/*
 * delete_par_list() wipes out the contents of a raw_par array
 */
void
delete_par_list(int nr_par, struct raw_par *par)
{
    int x;

    for (x=0; x < nr_par; x++)
	free(par[x].par_dev);
    free(par);
} /* delete_par_list */


/*
 * get_disk_list() -- build a disk list containing all the disks we have NOT
 *                    autoconfigured.
 * This disklist is returned as a ListItem array, which the user may dispose
 * of with free() when they're done with it.
 */
LIA
get_disk_list(int rootdisk)
{
    int x;
    LIA list;

    list = newLIA(0,0);

    for (x=0; x < NrDisk; x++) {
	/* if this is the root device and we've autoconfigured it,
	 * don't add to the list.
	 */
	if (x == rootdisk)
	    continue;

	/* If it's doesn't exist or isn't a disk, don't add it */
	if (IS_OK(Disk[x].flags))
	    addToLIA(list, Disk[x].device, strdup(mediadesc(&Disk[x])), 0);
    }
    return list;
} /* get_disk_list */


/*
 * delete_disk_list() wipes out a previously defined disk list
 */
void
delete_disk_list(LIA disk_list)
{
    int x;

    for (x=0; x < LIAcount(disk_list); x++)
	free(LIAlist(disk_list)[x].item);
    free(LIAlist(disk_list));
} /* delete_disk_list */
